// @ts-nocheck
const rp = require('request-promise');
const { forEachLimit } = require('async');
const { writeFile } = require('fs');
const { fromCallback } = require('bluebird');
const cheerio = require('cheerio');
const ora = require('ora');
const spinner = ora('Amazon Scraper Started');
const { Parser } = require('json2csv');
const moment = require('moment');
const path = require('path'); 

const CONST = require('./constant');

class AmazonScraper {
    constructor({
        keyword,
        number,
        sponsored,
        proxy,
        cli,
        filetype,
        scrapeType,
        asin,
        sort,
        productsorttype,
        discount,
        rating,
        ua,
        timeout,
        randomUa,
        page,
        bulk,
        category,
        cookie,
        geo,
        asyncTasks,
        reviewFilter,
        rh,
        minPrice,
        maxPrice,
        minReviewCount,
        ignoredAsins,
    }) {
        this.asyncTasks = asyncTasks;
        this.asyncPage = 1;
        this.mainHost = `https://${geo.host}`;
        this.geo = geo;
        this.cookie = cookie;
        this.bulk = bulk;
        this.productSearchCategory = category;
        this.collector = [];
        this.keyword = keyword;
        this.number = parseInt(number, 10);
        this.searchPage = page;
        this.sponsored = sponsored;
        this.proxy = proxy;
        this.cli = cli;
        this.scrapeType = scrapeType;
        this.asin = asin;
        this.sort = sort;
        this.productsorttype = productsorttype;
        this.discount = discount;
        this.rating = rating;
        this.minRating = 1;
        this.maxRating = 5;
        this.timeout = timeout;
        this.randomUa = randomUa;
        this.totalProducts = 0;
        this.fileType = filetype;
        this.jsonToCsv = new Parser({ flatten: true });
        this.initTime = Date.now();
        this.ua = ua || 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/81.0.4044.113 Safari/537.36';
        this.reviewFilter = reviewFilter;
        this.rh = rh
        this.minPrice = minPrice
        this.maxPrice = maxPrice
        this.minReviewCount = minReviewCount
        
        if (ignoredAsins) {
            this.ignoredAsins = ignoredAsins.toLowerCase().split(',')
        } else { this.ignoredAsins = null }

        if (this.proxy) this.proxy = this.proxy.split(',')
    }

    /**
     * Get user agent
     * if randomUa then user agent version will be randomized, this helps to prevent request blocking from the amazon side
     */
    get userAgent() {
        return this.randomUa
            ? `Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/${
                  Math.floor(Math.random() * 14) + 65
              }.0.4044.113 Safari/537.36`
            : this.ua;
    }

    get getProxy() {
        if (Array.isArray(this.proxy)) {
            return this.proxy[Math.floor(Math.random() * this.proxy.length)];
        }

        return '';
    }
    /**
     * Main request method
     * @param {*} param0
     */
    httpRequest({ uri, headers, method, qs, json, body, form }) {
        return new Promise(async (resolve, reject) => {
            const options = {
                uri: uri ? `${this.mainHost}/${uri}` : this.mainHost,
                method,
                ...(qs ? { qs } : {}),
                ...(body ? { body } : {}),
                ...(form ? { form } : {}),
                headers: {
                    'user-agent': this.userAgent,
                    ...headers,
                    accept: 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
                    'accept-language': 'en-US,en;q=0.5',
                    'accept-encoding': 'gzip, deflate, br',
                    te: 'trailers',
                },
                ...(json ? { json: true } : {}),
                gzip: true,
                resolveWithFullResponse: true,
                ...(this.getProxy ? { proxy: `http://${this.getProxy}/` } : {}),
            };
            try {
                const response = await rp(options);
                setTimeout(() => {
                    resolve(response);
                }, this.timeout);
            } catch (error) {
                reject(error);
            }
        });
    }

    /**
     * Start scraper
     */
    async startScraper() {
        if (this.scrapeType === 'products') {
            this.asyncPage = Math.ceil(this.number / 15);
            if (!this.keyword) {
                throw new Error('Keyword is missing');
            }
            if (this.number > CONST.limit.product) {
                throw new Error(`Wow.... slow down cowboy. Maximum you can get is ${CONST.limit.product} products`);
            }
            if (typeof this.sponsored !== 'boolean') {
                throw new Error('Sponsored can only be {true} or {false}');
            }
        }
        if (this.scrapeType === 'reviews') {
            this.asyncPage = Math.ceil(this.number / 10);
            if (!this.asin) {
                throw new Error('ASIN is missing');
            }
            if (this.number > CONST.limit.reviews) {
                throw new Error(`Wow.... slow down cowboy. Maximum you can get is ${CONST.limit.reviews} reviews`);
            }
        }
        if (this.scrapeType === 'asin') {
            if (!this.asin) {
                throw new Error('ASIN is missing');
            }
        }
        if (!Array.isArray(this.rating)) {
            throw new Error('rating can only be an array with length of 2');
        }

        if (this.rating.length > 2) {
            throw new Error('rating can only be an array with length of 2');
        }

        if (!parseFloat(this.rating[0]) || !parseFloat(this.rating[1])) {
            throw new Error('rating can only contain 2 float values');
        }

        this.minRating = parseFloat(this.rating[0]);

        this.maxRating = parseFloat(this.rating[1]);

        if (this.minRating > this.maxRating) {
            throw new Error(`min rating can't be larger then max rating`);
        }
        if (this.cli) {
            spinner.start();
        }

        await this.mainLoop();
        this.sortAndFilterResult();

        const filePaths = await this.saveResultToFile();

        if (this.cli) {
            spinner.stop();
        }

        if (this.fileType && this.cli) {
            if (filePaths) {
                console.log('Result was saved to:', filePaths);
            } else {
                console.log(`Could not save results`);
            }
        }
        return {
            ...(this.scrapeType === 'products' ? { totalProducts: this.totalProducts, category: this.productSearchCategory } : {}),
            result: this.collector,
        };
    }

    /**
     * Main loop that collects data
     */
    async mainLoop() {
        return new Promise((resolve) => {
            forEachLimit(
                Array.from({ length: this.asyncPage }, (_, k) => k + 1),
                this.asyncTasks,
                async (item) => {
                    const body = await this.buildRequest(this.bulk ? item : this.searchPage);
                    if (this.scrapeType === 'asin') {
                        this.grapAsinDetails(body);
                        throw new Error('Done');
                    }
                    if (this.scrapeType === 'products') {
                        let totalResultCount = body.match(/"totalResultCount":\w+(.[0-9])/gm);

                        if (totalResultCount) {
                            this.totalProducts = totalResultCount[0].split('totalResultCount":')[1];
                        }
                        this.grabProduct(body, item);
                    }
                    if (this.scrapeType === 'reviews') {
                        this.grabReviews(body);
                    }
                    if (!this.bulk) {
                        throw new Error('Done');
                    }
                },
                () => {
                    resolve();
                },
            );
        });
    }

    /**
     * Get filename
     */
    get fileName() {
        switch (this.scrapeType) {
            case 'products':
                return `${this.scrapeType}-${this.keyword.replace(' ', '_')}-${this.initTime}`;
            case 'reviews':
            case 'asin':
                return `${this.scrapeType}-${this.asin}-${this.initTime}`;
            default:
                throw new Error(`Unknow scraping type: ${this.scrapeType}`);
        }
    }
    /**
     * Save results to the file
     */
    async saveResultToFile() {
        var filePaths = []

        if (this.collector.length) {
            var fileTypes = []

            switch (this.fileType) {
                case 'json':
                    fileTypes = ['json']
                    break;
                case 'csv':
                    fileTypes = ['csv']
                    break;
                case 'all':
                    fileTypes = ['json', 'csv']
                    break;
                default:
                    break;
            }

            for (const fileType of fileTypes) {
                const filePath = `${this.fileName}.${fileType}`
                await fromCallback((cb) => writeFile(
                    filePath,
                    fileType === 'json'
                        ? JSON.stringify(this.collector)
                        : this.jsonToCsv.parse(this.collector),
                    cb)
                );

                filePaths.push(path.resolve(filePath))
            }
        }

        return filePaths
    }

    /**
     * Sort and filet the result data
     */
    sortAndFilterResult() {
        if (this.scrapeType === 'reviews') {
            if (this.sort) {
                this.collector.sort((a, b) => b.rating - a.rating);
            }
        }
        if (this.scrapeType === 'products') {
            this.collector.sort((a, b) => a.position.global_position - b.position.global_position);
            this.collector.forEach((item, index) => {
                item.position.global_position = index += 1;
            });

            if (this.sort) {
                this.collector.sort((a, b) => b.score - a.score);
            }
            if (this.discount) {
                this.collector = this.collector.filter((item) => item.discounted);
            }
            if (this.sponsored) {
                this.collector = this.collector.filter((item) => item.sponsored);
            }
        }
        if (this.scrapeType === 'products' || this.scrapeType === 'reviews') {
            this.collector = this.collector.filter((item) => this.validateRating(item));
        }
    }

    get setRequestEndpoint() {
        switch (this.scrapeType) {
            case 'products':
                return 's';
            case 'reviews':
                return `product-reviews/${this.asin}/ref=cm_cr_arp_d_viewopt_srt?sortBy=${
                    CONST.reviewFilter.sortBy[this.reviewFilter.sortBy] ? CONST.reviewFilter.sortBy[this.reviewFilter.sortBy] : ''
                }${this.reviewFilter.verifiedPurchaseOnly ? '&reviewerType=avp_only_reviews' : ''}${
                    this.reviewFilter.filterByStar ? `&filterByStar=${CONST.reviewFilter.filterByStar[this.reviewFilter.filterByStar]}` : ''
                }`;
            case 'asin':
                return `dp/${this.asin}/ref=sspa_dk_detail_3&th=1&psc=1?th=1&psc=1`;
            default:
                return '';
        }
    }
    /**
     * Create request
     */
    async buildRequest(page) {
        const options = {
            method: 'GET',
            uri: this.setRequestEndpoint,
            qs: {
                ...(this.scrapeType === 'products'
                    ? {
                          k: this.keyword,
                          ...(this.productSearchCategory ? { i: this.productSearchCategory } : {}),
                          ...(this.productsorttype ? { s: this.productsorttype } : {}),
                          ...(this.rh ? { rh: this.rh } : {}),
                          ...(page > 1 ? { page, ref: `sr_pg_${page}` } : {}),
                      }
                    : {}),
                ...(this.scrapeType === 'reviews' ? { ...(page > 1 ? { pageNumber: page } : {}) } : {}),
            },
            headers: {
                referer: this.mainHost,
                cookie: this.cookie,
            },
        };

        try {
            const response = await this.httpRequest(options);
            return response.body;
        } catch (error) {
            throw error.message;
        }
    }

    /**
     * Collect reviews from the html response
     * @param {} body
     */
    grabReviews(body) {
        const $ = cheerio.load(body.replace(/\s\s+/g, '').replace(/\n/g, ''));
        const reviewsList = $('.a-section.a-spacing-none.review-views.celwidget')[0].children;
        let scrapingResult = {};
        for (let i = 0; i < reviewsList.length; i++) {
            const totalInResult = Object.keys(scrapingResult).length + this.collector.length;
            if (totalInResult >= this.number && this.bulk) {
                break;
            }
            const reviewId = reviewsList[i].attribs['id'];
            if (!reviewId) {
                continue;
            }
            scrapingResult[reviewId] = { id: reviewId };
        }
        /**
         * Review date
         */
        for (let key in scrapingResult) {
            const search = $(`#${key} [data-hook="review-date"]`);

            try {
                scrapingResult[key].review_data = search[0].children[0].data;
            } catch (error) {
                continue;
            }
        }

        /**
         * Reviewer name
         */
        for (let key in scrapingResult) {
            const search = $(`#${key} .a-profile-name`);

            try {
                scrapingResult[key].name = search[0].children[0].data;
            } catch (error) {
                continue;
            }
        }

        /**
         * Rating
         */
        for (let key in scrapingResult) {
            const search = $(`#${key} [data-hook="review-star-rating"]`);

            try {
                scrapingResult[key].rating = parseFloat(search[0].children[0].children[0].data.split(' ')[0]);
            } catch (error) {
                continue;
            }
        }

        /**
         * Review title
         */
        for (let key in scrapingResult) {
            const search = $(`#${key} [data-hook="review-title"]`);

            try {
                scrapingResult[key].title = $(search[0]).text().toString();
            } catch (error) {
                continue;
            }
        }

        /**
         * Review text
         */
        for (let key in scrapingResult) {
            const search = $(`#${key} [data-hook="review-body"]`);

            try {
                scrapingResult[key].review = $(search[0]).text();
            } catch (error) {
                continue;
            }
        }

        /**
         * If purchase is verified
         */
        for (let key in scrapingResult) {
            const search = $(`#${key} [data-reftag="cm_cr_arp_d_rvw_rvwer"]`);
            scrapingResult[key].verified_purchase = false;

            try {
                scrapingResult[key].verified_purchase = !!search[0];
            } catch (error) {
                continue;
            }
        }

        for (let key in scrapingResult) {
            this.collector.push(scrapingResult[key]);
        }
    }

    // Extract product features
    extractProductFeatures($) {
        const featureBullets = [];

        try {
            const featureList = $('#feature-bullets > ul .a-list-item');
            for (let item in featureList) {
                if (featureList[item].children[0].data) featureBullets.push(featureList[item].children[0].data);
            }
        } catch (error) {
            //continue regardless of error
        }

        // Features on some items can be hidden with the expander tag
        // We will try to extract them
        try {
            const featureListExpanded = $('#feature-bullets > div > div')[0].children[0].children;
            for (let item of featureListExpanded) {
                if (item.children[0].children[0].data) featureBullets.push(item.children[0].children[0].data);
            }
        } catch (error) {
            //continue regardless of error
        }

        return featureBullets;
    }

    // Extract infromation from the Product Information section
    extractProductInfromation($) {
        const bestsellers_rank = [];
        const product_information = {
            dimensions: '',
            weight: '',
            available_from: '',
            available_from_utc: '',
            available_for_months: 0,
            available_for_days: 0,
            manufacturer: '',
            model_number: '',
            department: '',
        };

        let unOrderedList = false;

        try {
            const bestSalesSection = $('#SalesRank > ul > li');
            for (let item in bestSalesSection) {
                const rank = $(bestSalesSection[item].children[0]).text();
                const category = $(bestSalesSection[item].children[1]).text();
                const link = bestSalesSection[item].children[1].children[1].attribs.href;
                if (category && rank) {
                    const bestSeller = this.geo.best_seller(`${rank} ${category}`);
                    if (bestSeller) {
                        bestsellers_rank.push({ ...bestSeller, link: `${this.mainHost}${link}` });
                    }
                }
            }
        } catch (error) {}

        try {
            const bestSalesSection = $('#detailBulletsWrapper_feature_div > ul:nth-child(5) > li > span > ul > li');
            for (let item in bestSalesSection) {
                const rank = $(bestSalesSection[item].children[0].children[0]).text();
                const category = $(bestSalesSection[item].children[0].children[1]).text();
                const link = bestSalesSection[item].children[0].children[1].attribs.href;
                if (category && rank) {
                    const bestSeller = this.geo.best_seller(`${rank} ${category}`);
                    if (bestSeller) {
                        bestsellers_rank.push({ ...bestSeller, link: `${this.mainHost}${link}` });
                    }
                }
            }
        } catch (error) {}

        try {
            for (let id of this.geo.product_information.id) {
                const informationSection = $(id);
                if (informationSection[0]) {
                    const len = informationSection[0].children.length;
                    let arr = informationSection[0].children;
                    if (len === 1) {
                        arr = informationSection[0].children[0].children;
                    }

                    for (let item of arr) {
                        let key = '';
                        let value = '';
                        if (len === 1) {
                            try {
                                // Some of the product information can be located in the unordered list of data
                                if (id === '#detailBulletsWrapper_feature_div > ul:nth-child(5)') {
                                    unOrderedList = true;
                                    key = item.children[0].data;
                                    value = item.children[1].data;
                                } else {
                                    key = item.children[0].children[0].data;
                                    value = item.children[1].children[0].data;
                                }
                            } catch (error) {
                                //continue regardless of error
                            }
                        } else {
                            try {
                                key = item.children[0].children[0].children[0].data.replace(':', '');
                                value = item.children[0].children[1].children[0].data;
                            } catch (error) {
                                //continue regardless of error
                            }
                        }

                        const fieldToExtract = this.geo.product_information.fields[key];
                        if (fieldToExtract) {
                            // Extracting data from the Best Sellers Rank section
                            if (fieldToExtract.rank) {
                                try {
                                    let rankArray = [];
                                    if (unOrderedList) {
                                        rankArray = item.children[1].children;
                                    } else {
                                        rankArray = item.children[1].children[0].children;
                                    }
                                    rankArray.forEach((item) => {
                                        if (item.name === 'span') {
                                            const bestSeller = this.geo.best_seller($(item).text());
                                            const link = item.children[1].attribs.href;
                                            if (bestSeller) {
                                                bestsellers_rank.push({ ...bestSeller, link: `${this.mainHost}${link}` });
                                            }
                                        }
                                    });
                                } catch (error) {
                                    //continue regardless of error
                                }
                            }

                            if (fieldToExtract && value) {
                                if (fieldToExtract.key === 'dimensions') {
                                    // Sometimes weight is located in the dimensions section and in such cases we can separate both values and place them in the correct fields
                                    if (value.indexOf('; ') > -1) {
                                        const splitDimensions = value.split('; ');
                                        if (!product_information.weight && splitDimensions[1]) {
                                            product_information.weight = splitDimensions[1];
                                        }
                                        value = splitDimensions[0];
                                    }
                                }
                                product_information[fieldToExtract.key] = value;
                            }
                        }
                    }
                }
            }
        } catch (error) {
            //continue regardless of error
        }

        // Calculate months and dyas on amazon
        if (product_information.available_from) {
            const from = moment(new Date(product_information.available_from));
            const now = moment(new Date());
            const duration = moment.duration(now.diff(from));
            product_information.available_from_utc = new Date(product_information.available_from).toISOString();
            product_information.available_for_months = Math.ceil(duration.asMonths());
            product_information.available_for_days = Math.ceil(duration.asDays());
        }

        return [product_information, bestsellers_rank];
    }

    // Extract products from the variants section (different color,size and etc)
    extractProductVariants($, body) {
        const variants = [];
        let videos = [];

        let variantsSelector = $('#variation_color_name > ul');
        let productJsonMeta = '';

        if (!variantsSelector[0]) {
            variantsSelector = $('#variation_style_name > ul');
        }

        if (!variantsSelector[0]) {
            variantsSelector = $('#variation_size_name > ul');
        }

        // Most of the variant related data can be extracted from the JSON that is located right on the product page
        // But sometimes it isn't and for such cases we will use again cheerio
        try {
            const json = /jQuery.parseJSON\('([\w{}:;&,-.[\]\s+%\/()\/"|\\`']+)'\)/.exec(body);

            if (json) {
                productJsonMeta = JSON.parse(json[1].replace(/\\'/g, `'`));
            }
        } catch (error) {
            //continue regardless of error
        }

        // If there is a valid JSON
        if (productJsonMeta) {
            if (Object.keys(productJsonMeta.colorToAsin).length) {
                for (let item in productJsonMeta.colorToAsin) {
                    const images = productJsonMeta.colorImages[item] || [];

                    const variant = {
                        asin: productJsonMeta.colorToAsin[item].asin,
                        images: images,
                        title: item,
                        link: `https://www.amazon.com/dp/${productJsonMeta.colorToAsin[item].asin}/?th=1&psc=1`,
                        is_current_product: productJsonMeta.colorToAsin[item].asin === this.asin,
                        price: '',
                    };
                    try {
                        const price = $(`li[data-defaultasin="${productJsonMeta.colorToAsin[item].asin}"] div.twisterSlotDiv`).text();
                        if (price) variant.price = price;
                    } catch (error) {
                        //continue regardless of error
                    }
                    variants.push(variant);
                }
            }
            videos = productJsonMeta.videos;
            // If there is no valid JSON
        } else {
            try {
                for (let item of variantsSelector[0].children) {
                    const variant = {
                        asin: '',
                        images: '',
                        title: '',
                        link: ``,
                        is_current_product: false,
                        price: '',
                    };
                    if (item.attribs['data-defaultasin']) {
                        variant.asin = item.attribs['data-defaultasin'];
                    }
                    if (item.attribs['data-dp-url']) {
                        const findAsin = /dp\/(\w+)\//.exec(item.attribs['data-dp-url']);
                        if (findAsin) {
                            variant.asin = findAsin[1];
                        }
                    }

                    try {
                        const price = $(`li[id=${item.attribs['id']}] div.twisterSlotDiv`).text();
                        if (price) variant.price = price;
                        variant.images =
                            item.children[0].children[0].children[0].children[0].children[0].children[0].children[1].children[0].children[0].attribs.src;
                    } catch (error) {
                        //continue regardless of error
                    }
                    variant.title = item.attribs['title'] ? item.attribs['title'].split(this.geo.variants.split_text)[1] : ''; //item.children[0].children[0].children[0].children[0].children[0].children[0].children[1].children[0].children[0].attribs.alt;

                    variants.push({
                        ...variant,
                        images: variant.images ? [`${variant.images.split('._')[0]}._AC_SY879_.jpg`] : [],
                        link: `https://www.amazon.com/dp/${variant.asin}/?th=1&psc=1`,
                        is_current_product: variant.asin === this.asin,
                    });
                }
            } catch (error) {
                //continue regardless of error
            }
        }
        return [variants, videos];
    }

    // Extract product from section "Customers who bought this item also bought"
    extractAlsoBought($) {
        const also_bought = [];
        const bought = $(
            '#desktop-dp-sims_purchase-similarities-sims-feature > div > div.a-row.a-carousel-controls.a-carousel-row.a-carousel-has-buttons > div > div.a-carousel-col.a-carousel-center > div',
        );

        if (!bought[0]) {
            return also_bought;
        }

        try {
            if (!bought[0].children[0].children) return also_bought;
        } catch (error) {
            return also_bought;
        }
        for (let item of bought[0].children[0].children) {
            const product = {
                asin: '',
                title: '',
                image: [],
                url: ``,
                reviews: {
                    total_reviews: 0,
                    rating: 0,
                },
                price: {
                    symbol: this.geo.symbol,
                    currency: this.geo.currency,
                    current_price: 0,
                },
                badges: {
                    amazon_prime: false,
                },
            };
            try {
                const asin = JSON.parse(item.children[0].attribs['data-p13n-asin-metadata']);
                product.asin = asin.asin;
                product.url = `https://www.amazon.com/dp/${product.asin}/?th=1&psc=1`;
            } catch (error) {
                // continue regardless of error
            }

            try {
                const titleImg = item.children[0].children[0].children[0].children[0];
                if (titleImg) {
                    product.title = titleImg.attribs.alt;
                    product.image = [titleImg.attribs.src];
                }
            } catch (error) {
                // continue regardless of error
            }

            try {
                const review = $(item.children[0].children[1].children[1]).text();
                if (review) {
                    product.reviews.total_reviews = parseInt(review.replace(/[^\d]/g, ''), 10);
                }
            } catch (error) {
                // continue regardless of error
            }

            try {
                const rating = item.children[0].children[1].children[0].children[0];
                if (rating) {
                    const classRating = rating.attribs.class;
                    if (classRating.indexOf('a-star-') > -1) {
                        const splitRating = classRating.split('a-star-')[1];
                        product.reviews.rating = parseFloat(splitRating.replace('-', '.'));
                    }
                }
            } catch (error) {
                // continue regardless of error
            }

            try {
                const price = $(item.children[0].children[2]).text();
                if (price) {
                    product.price.current_price = this.geo.price_format(price);
                }
            } catch (error) {
                // continue regardless of error
            }
            also_bought.push(product);
        }
        return also_bought;
    }

    // Extract products from the section "Sponsored products related to this item"
    extractSponsoredProducts($) {
        const sponsored_products = [];
        const sponsored = $('#sp_detail-none > div:nth-child(3) > div > div > div.a-carousel-col.a-carousel-center > div');

        if (!sponsored[0]) {
            return sponsored_products;
        }

        try {
            if (!sponsored[0].children[0].children) return sponsored_products;
        } catch (error) {
            return sponsored_products;
        }

        for (let item of sponsored[0].children[0].children) {
            const product = {
                asin: '',
                title: '',
                image: [],
                url: ``,
                reviews: {
                    total_reviews: 0,
                    rating: 0,
                },
                price: {
                    symbol: this.geo.symbol,
                    currency: this.geo.currency,
                    current_price: 0,
                },
                badges: {
                    amazon_prime: false,
                },
            };

            product.asin = item.children[1].attribs['data-asin'];
            product.url = `https://www.amazon.com/dp/${product.asin}/?th=1&psc=1`;

            try {
                const titleImg = $(`#sp_detail-none_${product.asin} > a > img`);
                if (titleImg[0]) {
                    product.title = titleImg[0].attribs.alt;
                    product.image = [titleImg[0].attribs.src];
                }
            } catch (error) {}

            const review = $(`#sp_detail-none_${product.asin} > div:nth-child(6) > a > span`).text();
            if (review) {
                product.reviews.total_reviews = parseInt(review.replace(/[^\d]/g, ''), 10);
            }

            const price = $(`#sp_detail-none_${product.asin} > div.a-row.a-color-price > a:nth-child(1) > span`).text();
            if (price) {
                product.price.current_price = this.geo.price_format(price);
            }

            const prime = $(`#sp_detail-none_${product.asin} > div.a-row.a-color-price > a:nth-child(2) > span > i`);
            if (prime[0]) {
                if (prime[0].attribs.class.indexOf('a-icon-prime') > -1) {
                    product.badges.amazon_prime = true;
                }
            }

            const rating = $(`#sp_detail-none_${product.asin} > div:nth-child(6) > a > i`);
            if (rating[0]) {
                const classRating = rating[0].attribs.class;
                if (classRating.indexOf('a-star-') > -1) {
                    const splitRating = classRating.split('a-star-')[1];
                    product.reviews.rating = splitRating.indexOf('-') > -1 ? splitRating.replace('-', '.') : splitRating;
                }
            }

            sponsored_products.push(product);
        }
        return sponsored_products;
    }
    /**
     * Collect asin details from the html response
     * @param {*} body
     */
    grapAsinDetails(body) {
        body = body.replace(/\s\s+/g, '').replace(/\n/g, '');
        const $ = cheerio.load(body);

        const output = {
            title: '',
            description: '',
            feature_bullets: [],
            variants: [],
            asin: this.asin,
            url: `${this.mainHost}/dp/${this.asin}`,
            reviews: {
                total_reviews: 0,
                rating: 0,
                answered_questions: 0,
            },
            item_available: true,
            price: {
                symbol: this.geo.symbol,
                currency: this.geo.currency,
                current_price: 0,
                discounted: false,
                before_price: 0,
                savings_amount: 0,
                savings_percent: 0,
            },
            bestsellers_rank: [],
            main_image: '',
            total_images: 0,
            images: [],
            total_videos: 0,
            videos: [],
            delivery_message: '',
            product_information: {
                dimensions: '',
                weight: '',
                available_from: '',
                available_for_months: 0,
                available_for_days: 0,
                manufacturer: '',
                model_number: '',
                department: '',

                store_id: '',
                brand: '',
                sold_by: '',
                fulfilled_by: '',
                qty_per_order: 'na',
            },
            badges: {
                amazon_сhoice: false,
                amazon_prime: false,
            },
            sponsored_products: [],
            also_bought: [],
        };

        output.item_available = $('span.qa-availability-message').text() ? false : true;

        output.sponsored_products = this.extractSponsoredProducts($);
        output.also_bought = this.extractAlsoBought($);

        [output.variants, output.videos] = this.extractProductVariants($, body);
        output.total_videos = output.videos.length;

        output.description = $('#productDescription').text();

        if (body.indexOf('bookDescription_feature_div') > -1 && !output.description) {
            output.description = $('#bookDescription_feature_div').text();
            output.description = $(output.description).text() || '';
        }

        output.feature_bullets = this.extractProductFeatures($);
        [output.product_information, output.bestsellers_rank] = this.extractProductInfromation($);

        const deliveryMessage = $('#deliveryMessageMirId').text();
        output.delivery_message = deliveryMessage;

        const merchantInfo = $('div[id="merchant-info"]').text();
        if (merchantInfo) {
            try {
                if (merchantInfo.indexOf('Ships from and sold by ') > -1) {
                    const splitSoldBy = merchantInfo.split('Ships from and sold by ')[1];
                    output.product_information.sold_by = splitSoldBy.slice(0, -1);
                    output.product_information.fulfilled_by = splitSoldBy.slice(0, -1);
                } else {
                    const splitSoldBy = merchantInfo.split('Sold by ')[1].split(' and ');
                    output.product_information.sold_by = splitSoldBy[0];
                    output.product_information.fulfilled_by = splitSoldBy[1].split('Fulfilled by ')[1].slice(0, -1);
                }
            } catch (error) {
                //continue regardless of error
            }
        }

        const maxOrderQty = $('select[id="quantity"]');
        if (maxOrderQty) {
            try {
                output.product_information.qty_per_order = maxOrderQty[0] ? maxOrderQty[0].children.length : 'na';
            } catch (error) {
                //continue regardless of error
            }
        }

        output.title = $(`span[id="productTitle"]`).text() || $('.qa-title-text').text();
        output.reviews.total_reviews = $(`span[id="acrCustomerReviewText"]`).text().split(/\s/)[0]
            ? parseInt(
                  $(`span[id="acrCustomerReviewText"]`)
                      .text()
                      .split(/\s/)[0]
                      .replace(/[^0-9]/g, ''),
                  10,
              )
            : 0;
        output.reviews.answered_questions = $(`a[id="askATFLink"]`).text() ? parseInt($(`a[id="askATFLink"]`).text(), 10) : 0;
        output.reviews.rating = $(`span.reviewCountTextLinkedHistogram.noUnderline`)[0]
            ? $($(`span.reviewCountTextLinkedHistogram.noUnderline`)[0]).text().split(/\s/g)[0]
            : 0;

        output.price.current_price = $(`#priceblock_ourprice`)[0] ? this.geo.price_format($($(`#priceblock_ourprice`)[0].children[0]).text()) : 0;

        output.product_information.store_id = $(`input[id="storeID"]`).val();
        output.product_information.brand = $(`a[id="bylineInfo"]`).text() || '';

        output.badges.amazon_сhoice = $(`div.ac-badge-wrapper`)[0] ? true : false;
        output.badges.amazon_prime = $(`span[id="priceBadging_feature_div"]`)[0] ? true : false;

        output.price.discounted = $(`span.priceBlockStrikePriceString.a-text-strike`)[0] ? true : false;
        output.price.before_price = output.price.discounted
            ? this.geo.price_format($(`span.priceBlockStrikePriceString.a-text-strike`).text())
            : output.price.current_price;

        if (output.price.discounted) {
            output.price.savings_amount = +(output.price.before_price - output.price.current_price).toFixed(2);

            output.price.savings_percent = +((100 / output.price.before_price) * output.price.savings_amount).toFixed(2);
        }

        const thumbnail = $('span[data-action="thumb-action"]');
        for (let i = 0; i < thumbnail.length; i++) {
            try {
                let url = thumbnail[i].children[0].children[0].children[1].children[0].attribs.src;
                if (url.indexOf('x-locale/common') === -1) {
                    output.images.push(`${url.split('._')[0]}._AC_SY879_.jpg`);
                }
            } catch (error) {
                // continue regardless of error
            }
        }
        output.main_image = output.images.length ? output.images[0] : '';
        output.total_images = output.images.length;

        this.collector.push(output);
    }

    /**
     * Collect products from html response
     * @param {*} body
     */
    grabProduct(body, p) {
        const $ = cheerio.load(body.replace(/\s\s+/g, '').replace(/\n/g, ''));
        const productList = $('div[data-index]');
        const scrapingResult = {};

        if (productList.length < 10) {
            throw new Error('No more products');
        }
        let position = 0;
        for (let i = 0; i < productList.length; i++) {
            if (this.cli) {
                spinner.text = `Found ${this.collector.length + productList.length} products`;
            }

            const asin = productList[i].attribs['data-asin'];
            if (!asin) {
                continue;
            }

            scrapingResult[asin] = {
                position: {
                    page: p,
                    position: (position += 1),
                    global_position: `${p}${i}`,
                },
                asin,
                price: {
                    discounted: false,
                    current_price: 0,
                    currency: this.geo.currency,
                    before_price: 0,
                    savings_amount: 0,
                    savings_percent: 0,
                },
                reviews: {
                    total_reviews: 0,
                    rating: 0,
                },
                url: `${this.mainHost}/dp/${asin}`,
                score: 0,
                sponsored: false,
                amazonChoice: false,
                bestSeller: false,
                amazonPrime: false,
            };
        }

        for (let key in scrapingResult) {
            try {
                const priceSearch = $(`div[data-asin=${key}] span[data-a-size="l"]`)[0] || $(`div[data-asin=${key}] span[data-a-size="m"]`)[0];
                const discountSearch = $(`div[data-asin=${key}] span[data-a-strike="true"]`)[0];
                const ratingSearch = $(`div[data-asin=${key}] .a-icon-star-small`)[0];
                const titleThumbnailSearch = $(`div[data-asin=${key}] [data-image-source-density="1"]`)[0];
                const amazonChoice = $(`div[data-asin=${key}] span[id="${key}-amazons-choice"]`).text();
                const bestSeller = $(`div[data-asin=${key}] span[id="${key}-best-seller"]`).text();
                const amazonPrime = $(`div[data-asin=${key}] .s-prime`)[0];

                if (priceSearch) {
                    scrapingResult[key].price.current_price = this.geo.price_format($(priceSearch.children[0]).text());
                }

                if (amazonChoice) {
                    scrapingResult[key].amazonChoice = true;
                }
                if (bestSeller) {
                    scrapingResult[key].bestSeller = true;
                }
                if (amazonPrime) {
                    scrapingResult[key].amazonPrime = true;
                }

                if (discountSearch) {
                    scrapingResult[key].price.before_price = this.geo.price_format($(discountSearch.children[0]).text());

                    scrapingResult[key].price.discounted = true;

                    const savings = scrapingResult[key].price.before_price - scrapingResult[key].price.current_price;
                    if (savings <= 0) {
                        scrapingResult[key].price.discounted = false;

                        scrapingResult[key].price.before_price = 0;
                    } else {
                        scrapingResult[key].price.savings_amount = +(
                            scrapingResult[key].price.before_price - scrapingResult[key].price.current_price
                        ).toFixed(2);
                        scrapingResult[key].price.savings_percent = +(
                            (100 / scrapingResult[key].price.before_price) *
                            scrapingResult[key].price.savings_amount
                        ).toFixed(2);
                    }
                }

                if (ratingSearch) {
                    scrapingResult[key].reviews.rating = parseFloat(ratingSearch.children[0].children[0].data);

                    scrapingResult[key].reviews.total_reviews = parseInt(
                        ratingSearch.parent.parent.parent.next.attribs['aria-label'].replace(/\,/g, ''),
                    );

                    scrapingResult[key].score = parseFloat(scrapingResult[key].reviews.rating * scrapingResult[key].reviews.total_reviews).toFixed(2);
                }

                if (titleThumbnailSearch) {
                    scrapingResult[key].title = titleThumbnailSearch.attribs.alt;

                    scrapingResult[key].thumbnail = titleThumbnailSearch.attribs.src;
                }
            } catch (err) {
                continue;
            }
        }

        for (let key in scrapingResult) {
            this.collector.push(scrapingResult[key]);
        }
        return;
    }

    /**
     * Filter reviews/products by required rating
     * @param {*} item
     */
    validateRating(item) {
        if (this.scrapeType === 'products') {
            if ((item.reviews.rating >= this.minRating && item.reviews.rating <= this.maxRating)
                &&
                item.price.current_price
                &&
                item.asin && item.asin.length > 0
                &&
                (item.price.current_price && (!this.minPrice || item.price.current_price >= this.minPrice) && (!this.maxPrice || item.price.current_price <= this.maxPrice))
                &&
                (!this.minReviewCount || this.minReviewCount <= item.reviews.total_reviews)
                &&
                (!this.ignoredAsins || !this.ignoredAsins.includes(item.asin.toLowerCase()))
            ) {
                return item;
            }
        }
        if (this.scrapeType === 'reviews') {
            if (item.rating >= this.minRating && item.rating <= this.maxRating) {
                return item;
            }
        }
        return false;
    }

    async extractCategories() {
        const body = await this.buildRequest();
        const $ = cheerio.load(body.replace(/\s\s+/g, '').replace(/\n/g, ''));
        const categorySelect = $('#searchDropdownBox')[0];

        if (!Array.isArray(categorySelect.children)) {
            throw new Error("Can't find category selector");
        }

        let categories = {};
        for (let select of categorySelect.children) {
            const category = select.attribs.value.split('search-alias=')[1];
            categories[category] = {
                name: select.children[0].data,
                category,
            };
        }

        return categories;
    }
}

module.exports = AmazonScraper;
